#import "VNC.h"

@implementation VNCController

static VNCController *sharedInstance = nil;

+ (VNCController *)controller
{
    return sharedInstance;
}

+ (NSCursor *)hollowCursor
{
    return [sharedInstance hollowCursor];
}

- (NSCursor *)hollowCursor
{
    return _hollowCursor;
}

+ (void)initialize
{
    [[NSUserDefaults standardUserDefaults] registerDefaults:
        [NSDictionary dictionaryWithObjectsAndKeys:
            @"vncviewer.mydomain.org", @"VNCDefaultServerHost",
            @"1", @"VNCDefaultServerNumber",
            @"2", @"VNCDefaultColorDepthIndex",
            
            nil]];
}

- (NSArray *)displayManagers
{
    return [[_displayManagers copy] autorelease];
}

- (id)init
{
    if (self = [super init]) {
        [[NSNotificationCenter defaultCenter] addObserver:self
		selector:@selector(displayWillClose:)
		name:VNCDisplayWillTerminateNotification
		object:nil];
        
        sharedInstance = self;
        _hollowCursor = [[NSCursor alloc] initWithImage:[NSImage imageNamed:@"VNCHollowCursor"]
                                                hotSpot:NSMakePoint(8.0, 8.0)];
    }

    return self;
}

- (void)applicationDidFinishLaunching:(NSNotification *)notification
{
    _displayManagers = [[NSMutableArray alloc] init];
    
    [displayEntryTextField setStringValue:[[NSUserDefaults standardUserDefaults] stringForKey:@"VNCDefaultServerHost"]];
    [displayEntryPortField setStringValue:[[NSUserDefaults standardUserDefaults] stringForKey:@"VNCDefaultServerNumber"]];

    [displayEntryColorDepth selectItemAtIndex:[[NSUserDefaults standardUserDefaults] integerForKey:@"VNCDefaultColorDepthIndex"]];

    [displayMenu setAutoenablesItems:NO];
    [updateMenuItem setEnabled:NO];
}

- (void)displayWillClose:(NSNotification *)notification
{
    VNCDisplayManager *manager = [notification object];

    [_displayManagers removeObject:manager];
    if ([_displayManagers count] == 0)
        [updateMenuItem setEnabled:NO];
}

- (BOOL)applicationShouldTerminate:(NSApplication *)sender
{
    VNCDisplayManager *manager;
    NSEnumerator *displayEnumerator = [_displayManagers objectEnumerator];

    while (manager = [displayEnumerator nextObject])
        if (![manager performClose])
            return NO;

    return YES;
}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    [_displayManagers release];

    [super dealloc];
}

- (void)openPanelOkClicked:(id)sender
{
    NSString *host;
    int port = 0;
    VNCDisplayManager *manager;

    host = [displayEntryTextField stringValue];
    port = [displayEntryPortField intValue];

    [displayOpenPanel close];
    [[NSUserDefaults standardUserDefaults] setObject:host forKey:@"VNCDefaultServerHost"];
    [[NSUserDefaults standardUserDefaults] setInteger:port forKey:@"VNCDefaultServerNumber"];

    NS_DURING {
        _rfbProto = [RFBProtocolAdaptor adaptorWithHost:host port:port];

        [_rfbProto readServerVersionMessage];
        [_rfbProto sendClientVersionMessage];

        if ([_rfbProto readServerAuthScheme] == RFBVNCAuthorization) {
            NSData *challenge = [_rfbProto readServerAuthChallenge];
            NSString *key = [NSString stringWithFormat:@"Password-%@-%d", host, port];
            NSString *password = [[NSUserDefaults standardUserDefaults] objectForKey:key];

            [authenticationPanelLabel setStringValue:[NSString stringWithFormat:@"Enter password for %@", host]];

            while (![_rfbProto authorizeWithPassword:password challenge:challenge]) {
                [NSApp runModalForWindow:authenticationPanel];
                [authenticationPanel orderOut:self];
                password = [authenticationPanelTextField stringValue];
            }

            if ([authenticationPanelCheckBox state]) {
                [[NSUserDefaults standardUserDefaults] setObject:password forKey:key];
            }
        }

        // do initial handshaking
        [_rfbProto sendClientInitializationWithSharedDesktop:YES];
        [_rfbProto readServerInitializationMessage];

        // set up transmission
        [_rfbProto sendEncodings:[RFBRectangle supportedEncodings]
                           count:[RFBRectangle numberOfSupportedEncodings]];

        // set up the VNC color spaces in a way that makes DPS happy
        switch ([displayEntryColorDepth indexOfSelectedItem]) {
            case 0:		// 24-bit color
                [_rfbProto sendBitsPerPixel:32 depth:24
                                  bigEndian:YES trueColour:YES
                                     redMax:255 greenMax:255 blueMax:255
                                   redShift:24 greenShift:16 blueShift:8];
                break;

            case 1:		// 12-bit color
                [_rfbProto sendBitsPerPixel:16 depth:12
                                  bigEndian:YES trueColour:YES
                                     redMax:15 greenMax:15 blueMax:15
                                   redShift:12 greenShift:8 blueShift:4];
                break;

            case 2:		// 8-bit color
                [_rfbProto sendBitsPerPixel:8 depth:6
                                  bigEndian:YES trueColour:YES
                                     redMax:3 greenMax:3 blueMax:3
                                   redShift:6 greenShift:4 blueShift:2];
                break;

            default:
                [NSException raise:NSInvalidArgumentException
                            format:@"broken interface - tag for color depth not 24, 12, or 8"];
        }
        [[NSUserDefaults standardUserDefaults] setInteger:[displayEntryColorDepth indexOfSelectedItem] forKey:@"VNCDefaultColorDepthIndex"];

        // once authorization is done, hand off to the display manager
        manager = [[[VNCDisplayManager alloc] initWithAdaptor:_rfbProto] autorelease];
        [_displayManagers addObject:manager];
        
        [updateMenuItem setEnabled:YES];
    }
    NS_HANDLER {
        NSRunAlertPanel(@"Connection error",
                        @"Exception connecting to %@:%d: %@",
                        @"Rats!", nil, nil,
                        host, port, [localException description]);
    }
    NS_ENDHANDLER;
}

- (void)authenticationPanelOkClicked:(id)sender
{
    [NSApp stopModal];
}

- (void)updateForemostDisplay:(id)sender
{
    VNCDisplayManager *manager = [[NSApp mainWindow] delegate];

    if (manager != nil) {
        if (NSRunAlertPanel(@"Confirm Update",
                            @"Updating \"%@\" may take a long time. Really update?",
                            @"Yes", @"No", nil,
                            [[manager adaptor] desktopName]) == NSAlertDefaultReturn)
            [manager requestFullUpdate:nil];
    }
}

- (void)orderFrontStatisticsPanel:(id)sender
{
}

- (void)takeSnapshotOfFrontWindow:(id)sender
{
    VNCDisplayManager *manager = [[NSApp mainWindow] delegate];

    if (manager != nil) {
        NSSavePanel *savePanel = [NSSavePanel savePanel];

        [savePanel setTitle:@"Save Desktop"];
        [savePanel setRequiredFileType:@"tiff"];

        if ([savePanel runModalForDirectory:@""
                                       file:[[manager adaptor] desktopName]] == NSOKButton)
            [[manager desktopTIFFRepresentation] writeToFile:[savePanel filename]
                                                  atomically:NO];
    }
}

@end
